/******************************************************************************
 * The MIT License
 *
 * Copyright (c) 2010 Perry Hung.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *****************************************************************************/

/**
 * @brief flash peripheral management functions
 */
#include "libmaple.h"
#include "flash.h"

/* flash registers  */
#define FLASH_BASE                      0x40022000
#define FLASH_ACR                       FLASH_BASE

/* flash prefetcher  */
#define ACR_PRFTBE                      BIT(4)
#define ACR_PRFTBE_ENABLE               BIT(4)

/* flash wait states  */
#define ACR_LATENCY                     (0x7)
#define FLASH_WRITE_ACR(val)            __write(FLASH_ACR, val)
#define FLASH_READ_ACR()                __read(FLASH_ACR)

/**
 * @brief turn on the hardware prefetcher
 */
void flash_enable_prefetch(void) {
    uint32 val = FLASH_READ_ACR();

    val |= ACR_PRFTBE_ENABLE;

    FLASH_WRITE_ACR(val);
}

/**
 * @brief set flash wait states
 * @param number of wait states
 */
void flash_set_latency(uint32 wait_states) {
    uint32 val = FLASH_READ_ACR();

    val &= ~ACR_LATENCY;
    val |= wait_states;

    FLASH_WRITE_ACR(val);
}

void setupFLASH() {
	/* configure the HSI oscillator */
	if ((pRCC->CR & 0x01) == 0x00) {
		uint32 rwmVal = pRCC->CR;
		rwmVal |= 0x01;
		pRCC->CR = rwmVal;
	}
	
	/* wait for it to come on */
	while ((pRCC->CR & 0x02) == 0x00) {}
}

uint8 flashErasePage(uint32 pageAddr) {
	uint32 rwmVal = GET_REG(FLASH_CR);
	rwmVal = FLASH_CR_PER;
	SET_REG(FLASH_CR,rwmVal);
	
	while (GET_REG(FLASH_SR) & FLASH_SR_BSY) {}
	SET_REG(FLASH_AR,pageAddr);
	SET_REG(FLASH_CR,FLASH_CR_START | FLASH_CR_PER);
	while (GET_REG(FLASH_SR) & FLASH_SR_BSY) {}
	
	/* todo: verify the page was erased */
	
	rwmVal = 0x00;
	SET_REG(FLASH_CR,rwmVal);
	
	return 1;
}

uint8 flashErasePages(uint32 pageAddr, uint16 n) {
	while (n-->0) {
		if (!flashErasePage(pageAddr+0x400*n)) {
			return 0;
		}
	}
	
	return 1;
}

uint8 flashWriteWord(uint32 addr, uint32 word) {
	volatile uint16 *flashAddr = (volatile uint16*)addr;
	volatile uint32 lhWord = (volatile uint32)word & 0x0000FFFF;
	volatile uint32 hhWord = ((volatile uint32)word & 0xFFFF0000)>>16;
	
	uint32 rwmVal = GET_REG(FLASH_CR);
	SET_REG(FLASH_CR,FLASH_CR_PG);
	
	/* apparently we need not write to FLASH_AR and can
     simply do a native write of a half word */
	while (GET_REG(FLASH_SR) & FLASH_SR_BSY) {}
	*(flashAddr+0x01) = (uint16)hhWord;
	while (GET_REG(FLASH_SR) & FLASH_SR_BSY) {}
	*(flashAddr) = (uint16)lhWord;
	while (GET_REG(FLASH_SR) & FLASH_SR_BSY) {}
	
	rwmVal &= 0xFFFFFFFE;
	SET_REG(FLASH_CR,rwmVal);
	
	/* verify the write */
	if (*(volatile uint32*)addr != word) {
		return 0;
	}
	
	return 1;
}

void flashLock(void) {
	/* take down the HSI oscillator? it may be in use elsewhere */
	
	/* ensure all FPEC functions disabled and lock the FPEC */
	SET_REG(FLASH_CR,0x00000080);
}

void flashUnlock(void) {
	/* unlock the flash */
	SET_REG(FLASH_KEYR,FLASH_KEY1);
	SET_REG(FLASH_KEYR,FLASH_KEY2);
}


